@@ -1,6 +1,7 @@
 """Constants that are useful when working with OCR files."""
+from typing import List
 
-__all__ = ['NETS_ID']
+__all__: List[str] = ['NETS_ID']
 
 
 #: Nets' data sender/recipient ID.
