@@ -1,20 +1,43 @@
 """The higher-level objects API."""
 
-import collections
 import datetime
+from collections import OrderedDict
 from decimal import Decimal
-from typing import Iterable, List, Mapping, Optional, Union
+from typing import TYPE_CHECKING, Callable, ClassVar, Iterable, List, Optional
+from typing import OrderedDict as OrderedDictType
+from typing import TypeVar, Union, cast
 
-import attr
-from attr.validators import instance_of, optional
+from attrs import Factory, define, field
+from attrs.validators import instance_of, optional
 
-import netsgiro
-import netsgiro.records
-from netsgiro.records import Record
+from netsgiro.converters import (
+    to_assignment_type,
+    to_avtalegiro_registration_type,
+    to_service_code,
+    to_transaction_type,
+)
+from netsgiro.enums import AssignmentType, ServiceCode, TransactionType
+from netsgiro.records import (
+    AssignmentEnd,
+    AssignmentStart,
+    AvtaleGiroAgreement,
+    TransactionAmountItem1,
+    TransactionAmountItem2,
+    TransactionAmountItem3,
+    TransactionRecord,
+    TransactionSpecification,
+    TransmissionEnd,
+    TransmissionStart,
+)
+from netsgiro.records import parse as records_parse
 from netsgiro.validators import str_of_length
+from netsgiro.validators import validate_due_date as _validate_due_date
 
+if TYPE_CHECKING:
+    from netsgiro.enums import AvtaleGiroRegistrationType
+    from netsgiro.records import Record
 
-__all__ = [
+__all__: List[str] = [
     'Transmission',
     'Assignment',
     'Agreement',
@@ -23,8 +46,15 @@ __all__ = [
     'parse',
 ]
 
+TransactionAmountItems = Union[
+    TransactionAmountItem1, TransactionAmountItem2, TransactionAmountItem3
+]
 
-@attr.s
+# Record or Record subclasses
+R = TypeVar('R', bound='Record')
+
+
+@define
 class Transmission:
     """Transmission is the top-level object.
 
@@ -32,37 +62,36 @@ class Transmission:
     multiple :class:`~netsgiro.Assignment` objects of various types.
     """
 
-    #: Data transmitters unique enumeration of the transmission. String of 7
-    #: digits.
-    number = attr.ib(validator=str_of_length(7))
+    #: Data transmitters unique enumeration of the transmission. String of 7 digits.
+    number: str = field(validator=str_of_length(7))
 
     #: Data transmitter's Nets ID. String of 8 digits.
-    data_transmitter = attr.ib(validator=str_of_length(8))
+    data_transmitter: str = field(validator=str_of_length(8))
 
     #: Data recipient's Nets ID. String of 8 digits.
-    data_recipient = attr.ib(validator=str_of_length(8))
+    data_recipient: str = field(validator=str_of_length(8))
 
     #: For OCR Giro files from Nets, this is Nets' processing date.
     #:
     #: For AvtaleGiro payment request, the earliest due date in the
     #: transmission is automatically used.
-    date = attr.ib(default=None, validator=optional(instance_of(datetime.date)))
+    date: Optional[datetime.date] = field(
+        default=None, validator=optional(instance_of(datetime.date))
+    )
 
     #: List of assignments.
-    assignments = attr.ib(default=attr.Factory(list), repr=False)
+    assignments: List['Assignment'] = field(default=Factory(list), repr=False)
 
     @classmethod
-    def from_records(cls, records: List[Record]) -> 'Transmission':
+    def from_records(cls, records: List[R]) -> 'Transmission':
         """Build a Transmission object from a list of record objects."""
         if len(records) < 2:
-            raise ValueError(
-                'At least 2 records required, got {}'.format(len(records))
-            )
+            raise ValueError(f'At least 2 records required, got {len(records)}')
 
         start, body, end = records[0], records[1:-1], records[-1]
 
-        assert isinstance(start, netsgiro.records.TransmissionStart)
-        assert isinstance(end, netsgiro.records.TransmissionEnd)
+        assert isinstance(start, TransmissionStart)
+        assert isinstance(end, TransmissionEnd)
 
         return cls(
             number=start.transmission_number,
@@ -73,20 +102,18 @@ class Transmission:
         )
 
     @staticmethod
-    def _get_assignments(records: List[Record]) -> List['Assignment']:
-        assignments = collections.OrderedDict()
+    def _get_assignments(records: List[R]) -> List['Assignment']:
+        assignments: OrderedDictType[str, List[R]] = OrderedDict()
 
         current_assignment_number = None
         for record in records:
-            if isinstance(record, netsgiro.records.AssignmentStart):
+            if isinstance(record, AssignmentStart):
                 current_assignment_number = record.assignment_number
                 assignments[current_assignment_number] = []
             if current_assignment_number is None:
-                raise ValueError(
-                    'Expected AssignmentStart record, got {!r}'.format(record)
-                )
+                raise ValueError(f'Expected AssignmentStart record, got {record!r}')
             assignments[current_assignment_number].append(record)
-            if isinstance(record, netsgiro.records.AssignmentEnd):
+            if isinstance(record, AssignmentEnd):
                 current_assignment_number = None
 
         return [Assignment.from_records(rs) for rs in assignments.values()]
@@ -96,41 +123,47 @@ class Transmission:
         lines = [record.to_ocr() for record in self.to_records()]
         return '\n'.join(lines)
 
-    def to_records(self) -> Iterable[Record]:
+    def to_records(self) -> Iterable['Record']:
         """Convert the transmission to a list of records."""
         yield self._get_start_record()
         for assignment in self.assignments:
             yield from assignment.to_records()
         yield self._get_end_record()
 
-    def _get_start_record(self) -> Record:
-        return netsgiro.records.TransmissionStart(
-            service_code=netsgiro.ServiceCode.NONE,
+    def _get_start_record(self) -> 'Record':
+        return TransmissionStart(
+            service_code=ServiceCode.NONE,
             transmission_number=self.number,
             data_transmitter=self.data_transmitter,
             data_recipient=self.data_recipient,
         )
 
-    def _get_end_record(self) -> Record:
+    def _get_end_record(self) -> 'Record':
         avtalegiro_payment_request = all(
-            assignment.service_code == netsgiro.ServiceCode.AVTALEGIRO
+            assignment.service_code == ServiceCode.AVTALEGIRO
             and assignment.type
             in (
-                netsgiro.AssignmentType.TRANSACTIONS,
-                netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS,
+                AssignmentType.TRANSACTIONS,
+                AssignmentType.AVTALEGIRO_CANCELLATIONS,
             )
             for assignment in self.assignments
         )
         if self.assignments and avtalegiro_payment_request:
-            date = min(
-                assignment.get_earliest_transaction_date()
-                for assignment in self.assignments
-            )
+            date: Optional[datetime.date] = None
+            for assignment in self.assignments:
+                earliest_transaction_date = assignment.get_earliest_transaction_date()
+                if date is None or (
+                    isinstance(earliest_transaction_date, datetime.date)
+                    and earliest_transaction_date < date
+                ):
+                    date = earliest_transaction_date
         else:
             date = self.date
 
-        return netsgiro.records.TransmissionEnd(
-            service_code=netsgiro.ServiceCode.NONE,
+        assert isinstance(date, datetime.date)
+
+        return TransmissionEnd(
+            service_code=ServiceCode.NONE,
             num_transactions=self.get_num_transactions(),
             num_records=self.get_num_records(),
             total_amount=int(self.get_total_amount() * 100),
@@ -140,15 +173,14 @@ class Transmission:
     def add_assignment(
         self,
         *,
-        service_code: netsgiro.ServiceCode,
-        assignment_type: netsgiro.AssignmentType,
+        service_code: ServiceCode,
+        assignment_type: AssignmentType,
         agreement_id: Optional[str] = None,
         number: str,
         account: str,
-        date: Optional[datetime.date] = None
+        date: Optional[datetime.date] = None,
     ) -> 'Assignment':
-        """Add an assignment to the tranmission."""
-
+        """Add an assignment to the transmission."""
         assignment = Assignment(
             service_code=service_code,
             type=assignment_type,
@@ -162,27 +194,28 @@ class Transmission:
 
     def get_num_transactions(self) -> int:
         """Get number of transactions in the transmission."""
-        return sum(
-            assignment.get_num_transactions() for assignment in self.assignments
-        )
+        return sum(assignment.get_num_transactions() for assignment in self.assignments)
 
     def get_num_records(self) -> int:
         """Get number of records in the transmission.
 
         Includes the transmission's start and end record.
         """
-        return 2 + sum(
-            assignment.get_num_records() for assignment in self.assignments
-        )
+        return 2 + sum(assignment.get_num_records() for assignment in self.assignments)
 
     def get_total_amount(self) -> Decimal:
         """Get the total amount from all transactions in the transmission."""
-        return sum(
-            assignment.get_total_amount() for assignment in self.assignments
-        )
+        return Decimal(sum(assignment.get_total_amount() for assignment in self.assignments))
 
 
-@attr.s
+# Assigment transactions
+TS = List[Union['Transaction', 'Agreement', 'PaymentRequest']]
+
+# TransactionRecord and subclasses
+TR = TypeVar('TR', bound=TransactionRecord)
+
+
+@define
 class Assignment:
     """An Assignment groups multiple transactions within a transmission.
 
@@ -190,60 +223,64 @@ class Assignment:
     """
 
     #: The service code. One of :class:`~netsgiro.ServiceCode`.
-    service_code = attr.ib(converter=netsgiro.ServiceCode)
+    service_code: ServiceCode = field(converter=to_service_code)
 
-    #: The transaction type. One of :class:`~netsgiro.TransactionType`.
-    type = attr.ib(converter=netsgiro.AssignmentType)
+    #: The transaction type. One of :class:`~TransactionType`.
+    type: AssignmentType = field(converter=to_assignment_type)
 
     #: The assignment number. String of 7 digits.
-    number = attr.ib(validator=str_of_length(7))
+    number: str = field(validator=str_of_length(7))
 
     #: The payee's bank account. String of 11 digits.
-    account = attr.ib(validator=str_of_length(11))
+    account: str = field(validator=str_of_length(11))
 
     #: Used for OCR Giro.
     #:
     #: The payee's agreement ID with Nets. String of 9 digits.
-    agreement_id = attr.ib(default=None, validator=optional(str_of_length(9)))
+    agreement_id: Optional[str] = field(default=None, validator=optional(str_of_length(9)))
 
     #: Used for OCR Giro.
     #:
     #: The date the assignment was generated by Nets.
-    date = attr.ib(default=None, validator=optional(instance_of(datetime.date)))
+    date: Optional[datetime.date] = field(
+        default=None, validator=optional(instance_of(datetime.date))
+    )
 
     #: List of transaction objects, like :class:`~netsgiro.Agreement`,
     #: :class:`~netsgiro.PaymentRequest`, :class:`~netsgiro.Transaction`.
-    transactions = attr.ib(default=attr.Factory(list), repr=False)
+    transactions: TS = field(default=Factory(list), repr=False)
 
-    _next_transaction_number = 1
+    _next_transaction_number: int = field(default=1, init=False)
 
     @classmethod
-    def from_records(cls, records: List[Record]) -> 'Assignment':
+    def from_records(cls, records: List[R]) -> 'Assignment':
         """Build an Assignment object from a list of record objects."""
         if len(records) < 2:
-            raise ValueError(
-                'At least 2 records required, got {}'.format(len(records))
-            )
+            raise ValueError(f'At least 2 records required, got {len(records)}')
 
         start, body, end = records[0], records[1:-1], records[-1]
 
-        assert isinstance(start, netsgiro.records.AssignmentStart)
-        assert isinstance(end, netsgiro.records.AssignmentEnd)
+        assert isinstance(start, AssignmentStart)
+        assert isinstance(end, AssignmentEnd)
 
-        if start.service_code == netsgiro.ServiceCode.AVTALEGIRO:
-            if (
-                start.assignment_type
-                == netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS
-            ):
-                transactions = cls._get_agreements(body)
+        sc = start.service_code
+        at = start.assignment_type
+
+        cb: Callable
+        if sc == ServiceCode.OCR_GIRO:
+            # -> List[Transaction]
+            cb = cls._get_transactions
+        elif sc == ServiceCode.AVTALEGIRO:
+            if at == AssignmentType.AVTALEGIRO_AGREEMENTS:
+                # -> List[Agreement]
+                cb = cls._get_agreements
             else:
-                transactions = cls._get_payment_requests(body)
-        elif start.service_code == netsgiro.ServiceCode.OCR_GIRO:
-            transactions = cls._get_transactions(body)
-        else:
-            raise ValueError(
-                'Unknown service code: {}'.format(start.service_code)
-            )
+                # -> List[PaymentRequest]
+                cb = cls._get_payment_requests
+        else:  # pragma: no cover
+            raise ValueError(f'Unknown service code: {start.service_code}')
+
+        transactions: TS = cb(body)
 
         return cls(
             service_code=start.service_code,
@@ -256,43 +293,39 @@ class Assignment:
         )
 
     @staticmethod
-    def _get_agreements(records: List[Record]) -> List['Agreement']:
+    def _get_agreements(records: List[AvtaleGiroAgreement]) -> List['Agreement']:
         return [Agreement.from_records([r]) for r in records]
 
     @classmethod
-    def _get_payment_requests(
-        cls, records: List[Record]
-    ) -> List['PaymentRequest']:
+    def _get_payment_requests(cls, records: List[TR]) -> List['PaymentRequest']:
         transactions = cls._group_by_transaction_number(records)
         return [PaymentRequest.from_records(rs) for rs in transactions.values()]
 
     @classmethod
-    def _get_transactions(cls, records: List[Record]) -> List['Transaction']:
+    def _get_transactions(cls, records: List[TR]) -> List['Transaction']:
         transactions = cls._group_by_transaction_number(records)
         return [Transaction.from_records(rs) for rs in transactions.values()]
 
     @staticmethod
-    def _group_by_transaction_number(
-        records: List[Record],
-    ) -> Mapping[int, List[Record]]:
-        transactions = collections.OrderedDict()
-
-        for record in records:
-            if record.transaction_number not in transactions:
-                transactions[record.transaction_number] = []
-            transactions[record.transaction_number].append(record)
+    def _group_by_transaction_number(records: List[TR]) -> OrderedDictType[int, List[TR]]:
+        transactions: OrderedDictType[int, List[TR]] = OrderedDict()
+        for transaction_record in records:
+            tr = transaction_record.transaction_number
+            if tr not in transactions:
+                transactions[tr] = []
+            transactions[tr].append(transaction_record)
 
         return transactions
 
-    def to_records(self) -> Iterable[Record]:
+    def to_records(self) -> Iterable[Union[AssignmentStart, TransactionAmountItems, AssignmentEnd]]:
         """Convert the assignment to a list of records."""
         yield self._get_start_record()
         for transaction in self.transactions:
             yield from transaction.to_records()
         yield self._get_end_record()
 
-    def _get_start_record(self) -> Record:
-        return netsgiro.records.AssignmentStart(
+    def _get_start_record(self) -> AssignmentStart:
+        return AssignmentStart(
             service_code=self.service_code,
             assignment_type=self.type,
             assignment_number=self.number,
@@ -300,30 +333,28 @@ class Assignment:
             agreement_id=self.agreement_id,
         )
 
-    def _get_end_record(self) -> Record:
-        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
+    def _get_end_record(self) -> AssignmentEnd:
+        if self.service_code == ServiceCode.OCR_GIRO:
             dates = {
                 'nets_date_1': self.date,
                 'nets_date_2': self.get_earliest_transaction_date(),
                 'nets_date_3': self.get_latest_transaction_date(),
             }
-        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
+        elif self.service_code == ServiceCode.AVTALEGIRO:
             dates = {
                 'nets_date_1': self.get_earliest_transaction_date(),
                 'nets_date_2': self.get_latest_transaction_date(),
             }
-        else:
-            raise ValueError(
-                'Unhandled service code: {}'.format(self.service_code)
-            )
+        else:  # pragma: no cover
+            raise ValueError(f'Unhandled service code: {self.service_code}')
 
-        return netsgiro.records.AssignmentEnd(
+        return AssignmentEnd(
             service_code=self.service_code,
             assignment_type=self.type,
             num_transactions=self.get_num_transactions(),
             num_records=self.get_num_records(),
             total_amount=int(self.get_total_amount() * 100),
-            **dates
+            **dates,
         )
 
     def add_payment_request(
@@ -334,30 +365,30 @@ class Assignment:
         amount: Decimal,
         reference: Optional[str] = None,
         payer_name: Optional[str] = None,
-        bank_notification: Union[bool, str] = False
-    ) -> 'Transaction':
+        bank_notification: Union[bool, str] = False,
+        validate_due_date: bool = False,
+    ) -> 'PaymentRequest':
         """Add an AvtaleGiro payment request to the assignment.
 
         The assignment must have service code
-        :attr:`~netsgiro.ServiceCode.AVTALEGIRO` and assignment type
-        :attr:`~netsgiro.AssignmentType.TRANSACTIONS`.
+        :attr:`~ServiceCode.AVTALEGIRO` and assignment type
+        :attr:`~AssignmentType.TRANSACTIONS`.
         """
-
         assert (
-            self.service_code == netsgiro.ServiceCode.AVTALEGIRO
+            self.service_code == ServiceCode.AVTALEGIRO
         ), 'Can only add payment requests to AvtaleGiro assignments'
         assert (
-            self.type == netsgiro.AssignmentType.TRANSACTIONS
+            self.type == AssignmentType.TRANSACTIONS
         ), 'Can only add payment requests to transaction assignments'
 
+        if validate_due_date:
+            # Make sure we're not passing invalid due dates
+            _validate_due_date(due_date)
+
         if bank_notification:
-            transaction_type = (
-                netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION
-            )
+            transaction_type = TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION
         else:
-            transaction_type = (
-                netsgiro.TransactionType.AVTALEGIRO_WITH_PAYEE_NOTIFICATION
-            )
+            transaction_type = TransactionType.AVTALEGIRO_WITH_PAYEE_NOTIFICATION
 
         return self._add_avtalegiro_transaction(
             transaction_type=transaction_type,
@@ -377,27 +408,26 @@ class Assignment:
         amount: Decimal,
         reference: Optional[str] = None,
         payer_name: Optional[str] = None,
-        bank_notification: Union[bool, str] = False
-    ) -> 'Transaction':
+        bank_notification: Union[bool, str] = False,
+    ) -> 'PaymentRequest':
         """Add an AvtaleGiro cancellation to the assignment.
 
         The assignment must have service code
         :attr:`~netsgiro.ServiceCode.AVTALEGIRO` and assignment type
-        :attr:`~netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS`.
+        :attr:`~AssignmentType.AVTALEGIRO_CANCELLATIONS`.
 
         Otherwise, the cancellation must be identical to the payment request it
         is cancelling.
         """
-
         assert (
-            self.service_code == netsgiro.ServiceCode.AVTALEGIRO
+            self.service_code == ServiceCode.AVTALEGIRO
         ), 'Can only add cancellation to AvtaleGiro assignments'
         assert (
-            self.type == netsgiro.AssignmentType.AVTALEGIRO_CANCELLATIONS
+            self.type == AssignmentType.AVTALEGIRO_CANCELLATIONS
         ), 'Can only add cancellation to cancellation assignments'
 
         return self._add_avtalegiro_transaction(
-            transaction_type=netsgiro.TransactionType.AVTALEGIRO_CANCELLATION,
+            transaction_type=TransactionType.AVTALEGIRO_CANCELLATION,
             kid=kid,
             due_date=due_date,
             amount=amount,
@@ -409,20 +439,16 @@ class Assignment:
     def _add_avtalegiro_transaction(
         self,
         *,
-        transaction_type,
-        kid,
-        due_date,
-        amount,
-        reference=None,
-        payer_name=None,
-        bank_notification=None
-    ) -> 'Transaction':
-
-        if isinstance(bank_notification, str):
-            text = bank_notification
-        else:
-            text = ''
+        transaction_type: TransactionType,
+        kid: str,
+        due_date: datetime.date,
+        amount: Decimal,
+        reference: Optional[str] = None,
+        payer_name: Optional[str] = None,
+        bank_notification: Union[str, bool] = False,
+    ) -> 'PaymentRequest':
 
+        text = bank_notification if isinstance(bank_notification, str) else ''
         number = self._next_transaction_number
         self._next_transaction_number += 1
 
@@ -449,84 +475,73 @@ class Assignment:
 
         Includes the assignment's start and end record.
         """
-
-        return 2 + sum(
-            len(list(transaction.to_records()))
-            for transaction in self.transactions
-        )
+        return 2 + sum(len(list(transaction.to_records())) for transaction in self.transactions)
 
     def get_total_amount(self) -> Decimal:
         """Get the total amount from all transactions in the assignment."""
-        transactions = [
-            transaction
-            for transaction in self.transactions
-            if hasattr(transaction, 'amount')
-        ]
-        if not transactions:
-            return Decimal(0)
-        return sum(transaction.amount for transaction in transactions)
+        total = Decimal(0)
+        for t in self.transactions:
+            iter_amount = getattr(t, 'amount', None)
+            if iter_amount:
+                total += iter_amount
+        return total
+
+    def _get_earliest_or_latest_transaction_date(self, latest: bool) -> Optional[datetime.date]:
+        """Get earliest or latest date from the assignment's transactions."""
+        date_: Optional[datetime.date] = None
+        for t in self.transactions:
+            iter_date = getattr(t, 'date', None)
+            if iter_date and (
+                date_ is None or (iter_date > date_ if latest else iter_date < date_)
+            ):
+                date_ = iter_date
+        return date_
 
     def get_earliest_transaction_date(self) -> Optional[datetime.date]:
         """Get earliest date from the assignment's transactions."""
-        transactions = [
-            transaction
-            for transaction in self.transactions
-            if hasattr(transaction, 'date')
-        ]
-        if not transactions:
-            return None
-        return min(transaction.date for transaction in transactions)
+        return self._get_earliest_or_latest_transaction_date(latest=False)
 
     def get_latest_transaction_date(self) -> Optional[datetime.date]:
         """Get latest date from the assignment's transactions."""
-        transactions = [
-            transaction
-            for transaction in self.transactions
-            if hasattr(transaction, 'date')
-        ]
-        if not transactions:
-            return None
-        return max(transaction.date for transaction in transactions)
+        return self._get_earliest_or_latest_transaction_date(latest=True)
 
 
-@attr.s
+@define
 class Agreement:
     """Agreement contains an AvtaleGiro agreement update.
 
     Agreements are only found in assignments of the
-    :attr:`~netsgiro.AssignmentType.AVTALEGIRO_AGREEMENTS` type, which are only
+    :attr:`~AssignmentType.AVTALEGIRO_AGREEMENTS` type, which are only
     created by Nets.
     """
 
     #: The service code. One of :class:`~netsgiro.ServiceCode`.
-    service_code = attr.ib(converter=netsgiro.ServiceCode)
+    service_code: ServiceCode = field(converter=to_service_code)
 
     #: Transaction number. Unique and ordered within an assignment.
-    number = attr.ib(validator=instance_of(int))
+    number: int = field(validator=instance_of(int))
 
     #: Type of agreement registration update.
-    #: One of :class:`~netsgiro.AvtaleGiroRegistrationType`.
-    registration_type = attr.ib(converter=netsgiro.AvtaleGiroRegistrationType)
+    #: One of :class:`~AvtaleGiroRegistrationType`.
+    registration_type: 'AvtaleGiroRegistrationType' = field(
+        converter=to_avtalegiro_registration_type
+    )
 
     #: KID number to identify the customer and invoice.
-    kid = attr.ib(validator=optional(instance_of(str)))
+    kid: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: Whether the payer wants notification about payment requests.
-    notify = attr.ib(validator=instance_of(bool))
+    notify: bool = field(validator=instance_of(bool))
 
-    TRANSACTION_TYPE = netsgiro.TransactionType.AVTALEGIRO_AGREEMENT
+    TRANSACTION_TYPE: ClassVar[TransactionType] = TransactionType.AVTALEGIRO_AGREEMENT
 
     @classmethod
-    def from_records(cls, records: List[Record]) -> 'Agreement':
+    def from_records(cls, records: List[AvtaleGiroAgreement]) -> 'Agreement':
         """Build an Agreement object from a list of record objects."""
-
         assert len(records) == 1
         record = records[0]
-        assert isinstance(record, netsgiro.records.AvtaleGiroAgreement)
-        assert (
-            record.transaction_type
-            == netsgiro.TransactionType.AVTALEGIRO_AGREEMENT
-        )
+        assert isinstance(record, AvtaleGiroAgreement)
+        assert record.transaction_type == TransactionType.AVTALEGIRO_AGREEMENT
 
         return cls(
             service_code=record.service_code,
@@ -536,9 +551,9 @@ class Agreement:
             notify=record.notify,
         )
 
-    def to_records(self) -> Iterable[Record]:
+    def to_records(self) -> Iterable[AvtaleGiroAgreement]:
         """Convert the agreement to a list of records."""
-        yield netsgiro.records.AvtaleGiroAgreement(
+        yield AvtaleGiroAgreement(
             service_code=self.service_code,
             transaction_type=self.TRANSACTION_TYPE,
             transaction_number=self.number,
@@ -548,7 +563,7 @@ class Agreement:
         )
 
 
-@attr.s
+@define
 class PaymentRequest:
     """PaymentRequest contains an AvtaleGiro payment request or cancellation.
 
@@ -559,35 +574,35 @@ class PaymentRequest:
     """
 
     #: The service code. One of :class:`~netsgiro.ServiceCode`.
-    service_code = attr.ib(converter=netsgiro.ServiceCode)
+    service_code: ServiceCode = field(converter=to_service_code)
 
-    #: The transaction type. One of :class:`~netsgiro.TransactionType`.
-    type = attr.ib(converter=netsgiro.TransactionType)
+    #: The transaction type. One of :class:`~TransactionType`.
+    type: TransactionType = field(converter=to_transaction_type)
 
     #: Transaction number. Unique and ordered within an assignment.
-    number = attr.ib(validator=instance_of(int))
+    number: int = field(validator=instance_of(int))
 
     #: The due date.
-    date = attr.ib(validator=instance_of(datetime.date))
+    date: datetime.date = field(validator=instance_of(datetime.date))
 
     #: Transaction amount in NOK with two decimals.
-    amount = attr.ib(converter=Decimal)
+    amount: Decimal = field(converter=Decimal)
 
     #: KID number to identify the customer and invoice.
-    kid = attr.ib(validator=optional(instance_of(str)))
+    kid: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: This is a specification line that will, if set, be displayed on the
     #: payers account statement. Alphanumeric, max 25 chars.
-    reference = attr.ib(validator=optional(instance_of(str)))
+    reference: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: This is up to 42 lines of 80 chars each of free text used by the bank to
     #: notify the payer about the payment request. It is not used if the payee
     #: is responsible for notifying the payer.
-    text = attr.ib(validator=optional(instance_of(str)))
+    text: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: The value is only used to help the payee cross-reference reports from
-    #: Nets with their own records. It is not visible to the payer.
-    payer_name = attr.ib(validator=optional(instance_of(str)))
+    #: Nets with their own records. It is not vi.attr.ible to the payer.
+    payer_name: Optional[str] = field(validator=optional(instance_of(str)))
 
     @property
     def amount_in_cents(self) -> int:
@@ -595,18 +610,14 @@ class PaymentRequest:
         return int(self.amount * 100)
 
     @classmethod
-    def from_records(cls, records: List[Record]) -> 'Transaction':
+    def from_records(cls, records: List[TR]) -> 'PaymentRequest':
         """Build a Transaction object from a list of record objects."""
         amount_item_1 = records.pop(0)
-        assert isinstance(
-            amount_item_1, netsgiro.records.TransactionAmountItem1
-        )
+        assert isinstance(amount_item_1, TransactionAmountItem1)
         amount_item_2 = records.pop(0)
-        assert isinstance(
-            amount_item_2, netsgiro.records.TransactionAmountItem2
-        )
+        assert isinstance(amount_item_2, TransactionAmountItem2)
 
-        text = netsgiro.records.TransactionSpecification.to_text(records)
+        text = TransactionSpecification.to_text(cast(List['TransactionSpecification'], records))
 
         return cls(
             service_code=amount_item_1.service_code,
@@ -620,9 +631,11 @@ class PaymentRequest:
             payer_name=amount_item_2.payer_name,
         )
 
-    def to_records(self) -> Iterable[Record]:
+    def to_records(
+        self,
+    ) -> Iterable[Union[TransactionAmountItem1, TransactionAmountItem2, TransactionSpecification]]:
         """Convert the transaction to a list of records."""
-        yield netsgiro.records.TransactionAmountItem1(
+        yield TransactionAmountItem1(
             service_code=self.service_code,
             transaction_type=self.type,
             transaction_number=self.number,
@@ -630,7 +643,7 @@ class PaymentRequest:
             amount=self.amount_in_cents,
             kid=self.kid,
         )
-        yield netsgiro.records.TransactionAmountItem2(
+        yield TransactionAmountItem2(
             service_code=self.service_code,
             transaction_type=self.type,
             transaction_number=self.number,
@@ -638,10 +651,8 @@ class PaymentRequest:
             payer_name=self.payer_name,
         )
 
-        if self.type == (
-            netsgiro.TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION
-        ):
-            yield from netsgiro.records.TransactionSpecification.from_text(
+        if self.type == TransactionType.AVTALEGIRO_WITH_BANK_NOTIFICATION:
+            yield from TransactionSpecification.from_text(
                 service_code=self.service_code,
                 transaction_type=self.type,
                 transaction_number=self.number,
@@ -649,7 +660,7 @@ class PaymentRequest:
             )
 
 
-@attr.s
+@define
 class Transaction:
     """Transaction contains an OCR Giro transaction.
 
@@ -659,56 +670,54 @@ class Transaction:
     """
 
     #: The service code. One of :class:`~netsgiro.ServiceCode`.
-    service_code = attr.ib(converter=netsgiro.ServiceCode)
+    service_code: ServiceCode = field(converter=to_service_code)
 
-    #: The transaction type. One of :class:`~netsgiro.TransactionType`.
-    type = attr.ib(converter=netsgiro.TransactionType)
+    #: The transaction type. One of :class:`~TransactionType`.
+    type: TransactionType = field(converter=to_transaction_type)
 
     #: Transaction number. Unique and ordered within an assignment.
-    number = attr.ib(validator=instance_of(int))
+    number: int = field(validator=instance_of(int))
 
     #: Nets' processing date.
-    date = attr.ib(validator=instance_of(datetime.date))
+    date: datetime.date = field(validator=instance_of(datetime.date))
 
     #: Transaction amount in NOK with two decimals.
-    amount = attr.ib(converter=Decimal)
+    amount: Decimal = field(converter=Decimal)
 
     #: KID number to identify the customer and invoice.
-    kid = attr.ib(validator=optional(instance_of(str)))
+    kid: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: The value depends on the payment method.
-    reference = attr.ib(validator=optional(instance_of(str)))
+    reference: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: Up to 40 chars of free text from the payment terminal.
-    text = attr.ib(validator=optional(instance_of(str)))
+    text: Optional[str] = field(validator=optional(instance_of(str)))
 
     #: Used for OCR Giro.
-    centre_id = attr.ib(validator=optional(str_of_length(2)))
+    centre_id: Optional[str] = field(validator=optional(str_of_length(2)))
 
     #: Used for OCR Giro.
-    day_code = attr.ib(validator=optional(instance_of(int)))
+    day_code: Optional[int] = field(validator=optional(instance_of(int)))
 
     #: Used for OCR Giro.
-    partial_settlement_number = attr.ib(validator=optional(instance_of(int)))
+    partial_settlement_number: Optional[int] = field(validator=optional(instance_of(int)))
 
     #: Used for OCR Giro.
-    partial_settlement_serial_number = attr.ib(
-        validator=optional(str_of_length(5))
-    )
+    partial_settlement_serial_number: Optional[str] = field(validator=optional(str_of_length(5)))
 
     #: Used for OCR Giro.
-    sign = attr.ib(validator=optional(str_of_length(1)))
+    sign: Optional[str] = field(validator=optional(str_of_length(1)))
 
     #: Used for OCR Giro.
-    form_number = attr.ib(validator=optional(str_of_length(10)))
+    form_number: Optional[str] = field(validator=optional(str_of_length(10)))
 
     #: Used for OCR Giro.
-    bank_date = attr.ib(validator=optional(instance_of(datetime.date)))
+    bank_date: Optional[datetime.date] = field(validator=optional(instance_of(datetime.date)))
 
     #: Used for OCR Giro.
-    debit_account = attr.ib(validator=optional(str_of_length(11)))
+    debit_account: Optional[str] = field(validator=optional(str_of_length(11)))
 
-    _filler = attr.ib(validator=optional(str_of_length(7)))
+    _filler: Optional[str] = field(validator=optional(str_of_length(7)))
 
     @property
     def amount_in_cents(self) -> int:
@@ -716,20 +725,14 @@ class Transaction:
         return int(self.amount * 100)
 
     @classmethod
-    def from_records(cls, records: List[Record]) -> 'Transaction':
+    def from_records(cls, records: List[TR]) -> 'Transaction':
         """Build a Transaction object from a list of record objects."""
         amount_item_1 = records.pop(0)
-        assert isinstance(
-            amount_item_1, netsgiro.records.TransactionAmountItem1
-        )
+        assert isinstance(amount_item_1, TransactionAmountItem1)
         amount_item_2 = records.pop(0)
-        assert isinstance(
-            amount_item_2, netsgiro.records.TransactionAmountItem2
-        )
+        assert isinstance(amount_item_2, TransactionAmountItem2)
 
-        if len(records) == 1 and isinstance(
-            records[0], netsgiro.records.TransactionAmountItem3
-        ):
+        if len(records) == 1 and isinstance(records[0], TransactionAmountItem3):
             text = records[0].text
         else:
             text = None
@@ -746,9 +749,7 @@ class Transaction:
             centre_id=amount_item_1.centre_id,
             day_code=amount_item_1.day_code,
             partial_settlement_number=amount_item_1.partial_settlement_number,
-            partial_settlement_serial_number=(
-                amount_item_1.partial_settlement_serial_number
-            ),
+            partial_settlement_serial_number=amount_item_1.partial_settlement_serial_number,
             sign=amount_item_1.sign,
             form_number=amount_item_2.form_number,
             bank_date=amount_item_2.bank_date,
@@ -756,9 +757,9 @@ class Transaction:
             filler=amount_item_2._filler,
         )
 
-    def to_records(self) -> Iterable[Record]:
+    def to_records(self) -> Iterable[TransactionAmountItems]:
         """Convert the transaction to a list of records."""
-        yield netsgiro.records.TransactionAmountItem1(
+        yield TransactionAmountItem1(
             service_code=self.service_code,
             transaction_type=self.type,
             transaction_number=self.number,
@@ -768,12 +769,10 @@ class Transaction:
             centre_id=self.centre_id,
             day_code=self.day_code,
             partial_settlement_number=self.partial_settlement_number,
-            partial_settlement_serial_number=(
-                self.partial_settlement_serial_number
-            ),
+            partial_settlement_serial_number=self.partial_settlement_serial_number,
             sign=self.sign,
         )
-        yield netsgiro.records.TransactionAmountItem2(
+        yield TransactionAmountItem2(
             service_code=self.service_code,
             transaction_type=self.type,
             transaction_number=self.number,
@@ -785,10 +784,10 @@ class Transaction:
         )
 
         if self.type in (
-            netsgiro.TransactionType.REVERSING_WITH_TEXT,
-            netsgiro.TransactionType.PURCHASE_WITH_TEXT,
+            TransactionType.REVERSING_WITH_TEXT,
+            TransactionType.PURCHASE_WITH_TEXT,
         ):
-            yield netsgiro.records.TransactionAmountItem3(
+            yield TransactionAmountItem3(
                 service_code=self.service_code,
                 transaction_type=self.type,
                 transaction_number=self.number,
@@ -798,5 +797,4 @@ class Transaction:
 
 def parse(data: str) -> Transmission:
     """Parse an OCR file into a Transmission object."""
-    records = netsgiro.records.parse(data)
-    return Transmission.from_records(records)
+    return Transmission.from_records(records_parse(data))
