@@ -1,24 +1,47 @@
 """The lower-level records API."""
 
-import datetime
 import re
-from typing import Iterable, List, Optional, Sequence, Tuple, Union
+from abc import ABC, abstractmethod
+from typing import (
+    TYPE_CHECKING,
+    ClassVar,
+    Iterable,
+    List,
+    Optional,
+    Pattern,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 
-import attr
 from attr.validators import optional
+from attrs import define, field
 
-import netsgiro
+from netsgiro import RecordType, ServiceCode
 from netsgiro.converters import (
     fixed_len_str,
     stripped_newlines,
-    stripped_spaces_around,
-    truthy_or_none,
-    value_or_none,
+    to_assignment_type,
+    to_avtalegiro_registration_type,
+    to_bool,
+    to_date,
+    to_date_or_none,
+    to_int_or_none,
+    to_record_type,
+    to_safe_str_or_none,
+    to_service_code,
+    to_transaction_type,
 )
 from netsgiro.validators import str_of_length, str_of_max_length
 
+if TYPE_CHECKING:
+    import datetime
 
-__all__ = [
+    from netsgiro.enums import AssignmentType, AvtaleGiroRegistrationType, TransactionType
+
+__all__: List[str] = [
     'TransmissionStart',
     'TransmissionEnd',
     'AssignmentStart',
@@ -31,80 +54,34 @@ __all__ = [
     'parse',
 ]
 
-
-def to_service_code(
-    value: Union[netsgiro.ServiceCode, int, str]
-) -> netsgiro.ServiceCode:
-    return netsgiro.ServiceCode(int(value))
+R = TypeVar('R', bound='Record')
 
 
-def to_assignment_type(
-    value: Union[netsgiro.AssignmentType, int, str]
-) -> netsgiro.AssignmentType:
-    return netsgiro.AssignmentType(int(value))
+@define
+class Record(ABC):
+    """Record base class."""
 
+    _PATTERNS: ClassVar[List[Pattern]]
+    RECORD_TYPE: ClassVar[RecordType]
 
-def to_transaction_type(
-    value: Union[netsgiro.TransactionType, int, str]
-) -> netsgiro.TransactionType:
-    return netsgiro.TransactionType(int(value))
-
-
-def to_avtalegiro_registration_type(
-    value: Union[netsgiro.AvtaleGiroRegistrationType, int, str]
-) -> netsgiro.AvtaleGiroRegistrationType:
-    return netsgiro.AvtaleGiroRegistrationType(int(value))
-
-
-def to_date(value: Union[datetime.date, str, None]) -> Optional[datetime.date]:
-    if isinstance(value, datetime.date):
-        return value
-    if value is None or value == '000000':
-        return None
-    return datetime.datetime.strptime(value, '%d%m%y').date()
-
-
-def to_bool(value: Union[bool, str]) -> bool:
-    if isinstance(value, bool):
-        return value
-    if value == 'J':
-        return True
-    elif value == 'N':
-        return False
-    else:
-        raise ValueError("Expected 'J' or 'N', got {!r}".format(value))
-
-
-to_safe_str_or_none = value_or_none(
-    stripped_newlines(stripped_spaces_around(truthy_or_none(str)))
-)
-
-
-@attr.s
-class Record:
-    service_code = attr.ib(converter=to_service_code)
-
-    _PATTERNS = []
+    service_code: ServiceCode = field(converter=to_service_code)
 
     @classmethod
-    def from_string(cls, line: str) -> 'Record':
+    def from_string(cls: Type[R], line: str) -> R:
         """Parse OCR string into a record object."""
-
         for pattern in cls._PATTERNS:
             matches = pattern.match(line)
             if matches is not None:
                 return cls(**matches.groupdict())
 
-        raise ValueError(
-            '{!r} did not match {} record formats'.format(line, cls.__name__)
-        )
+        raise ValueError(f'{line!r} did not match {cls.__name__} record formats')
 
+    @abstractmethod
     def to_ocr(self) -> str:
         """Get record as OCR string."""
-        raise NotImplementedError
 
 
-@attr.s
+@define
 class TransmissionStart(Record):
     """TransmissionStart is the first record in every OCR file.
 
@@ -113,12 +90,12 @@ class TransmissionStart(Record):
     Each transmission can contain any number of assignments.
     """
 
-    transmission_number = attr.ib(validator=str_of_length(7))
-    data_transmitter = attr.ib(validator=str_of_length(8))
-    data_recipient = attr.ib(validator=str_of_length(8))
+    transmission_number: str = field(validator=str_of_length(7))
+    data_transmitter: str = field(validator=str_of_length(8))
+    data_recipient: str = field(validator=str_of_length(8))
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSMISSION_START
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSMISSION_START
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -141,24 +118,22 @@ class TransmissionStart(Record):
     def to_ocr(self) -> str:
         """Get record as OCR string."""
         return (
-            'NY000010'
-            '{self.data_transmitter:8}'
-            '{self.transmission_number:7}'
-            '{self.data_recipient:8}' + ('0' * 49)
-        ).format(self=self)
+            f'NY000010{self.data_transmitter:8}{self.transmission_number:7}{self.data_recipient:8}'
+            + ('0' * 49)
+        )
 
 
-@attr.s
+@define
 class TransmissionEnd(Record):
     """TransmissionEnd is the first record in every OCR file."""
 
-    num_transactions = attr.ib(converter=int)
-    num_records = attr.ib(converter=int)
-    total_amount = attr.ib(converter=int)
-    nets_date = attr.ib(converter=to_date)
+    num_transactions: int = field(converter=int)
+    num_records: int = field(converter=int)
+    total_amount: int = field(converter=int)
+    nets_date: 'datetime.date' = field(converter=to_date)
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSMISSION_END
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSMISSION_END
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -183,29 +158,30 @@ class TransmissionEnd(Record):
         """Get record as OCR string."""
         return (
             'NY000089'
-            '{self.num_transactions:08d}'
-            '{self.num_records:08d}'
-            '{self.total_amount:017d}'
-            '{self.nets_date:%d%m%y}' + ('0' * 33)
-        ).format(self=self)
+            f'{self.num_transactions:08d}'
+            f'{self.num_records:08d}'
+            f'{self.total_amount:017d}'
+            f'{self.nets_date:%d%m%y}'
+            + ('0' * 33)
+        )
 
 
-@attr.s
+@define
 class AssignmentStart(Record):
     """AssignmentStart is the first record of an assignment.
 
     Each assignment can contain any number of transactions.
     """
 
-    assignment_type = attr.ib(converter=to_assignment_type)
-    assignment_number = attr.ib(validator=str_of_length(7))
-    assignment_account = attr.ib(validator=str_of_length(11))
+    assignment_type: 'AssignmentType' = field(converter=to_assignment_type)
+    assignment_number: str = field(validator=str_of_length(7))
+    assignment_account: str = field(validator=str_of_length(11))
 
     # Only for assignment_type == AssignmentType.TRANSACTIONS
-    agreement_id = attr.ib(default=None, validator=optional(str_of_length(9)))
+    agreement_id: Optional[str] = field(default=None, validator=optional(str_of_length(9)))
 
-    RECORD_TYPE = netsgiro.RecordType.ASSIGNMENT_START
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.ASSIGNMENT_START
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -264,32 +240,29 @@ class AssignmentStart(Record):
     def to_ocr(self) -> str:
         """Get record as OCR string."""
         return (
-            'NY'
-            '{self.service_code:02d}'
-            '{self.assignment_type:02d}'
-            '20'
-            + (self.agreement_id and '{self.agreement_id:9}' or ('0' * 9))
-            + '{self.assignment_number:7}'
-            '{self.assignment_account:11}' + ('0' * 45)
-        ).format(self=self)
+            f'NY{self.service_code:02d}{self.assignment_type:02d}20'
+            + (self.agreement_id and f'{self.agreement_id:9}' or ('0' * 9))
+            + f'{self.assignment_number:7}{self.assignment_account:11}'
+            + ('0' * 45)
+        )
 
 
-@attr.s
+@define
 class AssignmentEnd(Record):
     """AssignmentEnd is the last record of an assignment."""
 
-    assignment_type = attr.ib(converter=to_assignment_type)
-    num_transactions = attr.ib(converter=int)
-    num_records = attr.ib(converter=int)
+    assignment_type: 'AssignmentType' = field(converter=to_assignment_type)
+    num_transactions: int = field(converter=int)
+    num_records: int = field(converter=int)
 
     # Only for transactions and cancellations
-    total_amount = attr.ib(default=None, converter=value_or_none(int))
-    nets_date_1 = attr.ib(default=None, converter=to_date)
-    nets_date_2 = attr.ib(default=None, converter=to_date)
-    nets_date_3 = attr.ib(default=None, converter=to_date)
+    total_amount: Optional[int] = field(default=None, converter=to_int_or_none)
+    nets_date_1: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
+    nets_date_2: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
+    nets_date_3: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
 
-    RECORD_TYPE = netsgiro.RecordType.ASSIGNMENT_END
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.ASSIGNMENT_END
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -348,91 +321,82 @@ class AssignmentEnd(Record):
     ]
 
     @property
-    def nets_date(self):
+    def nets_date(self) -> Optional['datetime.date']:
         """Nets' processing date.
 
         Only used for OCR Giro.
         """
-        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
-            return self.nets_date_1
-        else:
-            return None
+        return self.nets_date_1 if self.service_code == ServiceCode.OCR_GIRO else None
 
     @property
-    def nets_date_earliest(self):
+    def nets_date_earliest(self) -> Optional['datetime.date']:
         """Earliest date from the contained transactions."""
-
-        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
+        if self.service_code == ServiceCode.OCR_GIRO:
             return self.nets_date_2
-        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
+        elif self.service_code == ServiceCode.AVTALEGIRO:
             return self.nets_date_1
-        else:
-            raise ValueError(
-                'Unhandled service code: {}'.format(self.service_code)
-            )
+        else:  # pragma: no cover
+            raise ValueError(f'Unhandled service code: {self.service_code}')
 
     @property
-    def nets_date_latest(self):
+    def nets_date_latest(self) -> Optional['datetime.date']:
         """Latest date from the contained transactions."""
-
-        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
+        if self.service_code == ServiceCode.OCR_GIRO:
             return self.nets_date_3
-        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
+        elif self.service_code == ServiceCode.AVTALEGIRO:
             return self.nets_date_2
-        else:
-            raise ValueError(
-                'Unhandled service code: {}'.format(self.service_code)
-            )
+        else:  # pragma: no cover
+            raise ValueError(f'Unhandled service code: {self.service_code}')
 
     def to_ocr(self) -> str:
         """Get record as OCR string."""
         return (
             'NY'
-            '{self.service_code:02d}'
-            '{self.assignment_type:02d}'
+            f'{self.service_code:02d}'
+            f'{self.assignment_type:02d}'
             '88'
-            '{self.num_transactions:08d}'
-            '{self.num_records:08d}'
-            + (self.total_amount and '{self.total_amount:017d}' or ('0' * 17))
-            + (self.nets_date_1 and '{self.nets_date_1:%d%m%y}' or ('0' * 6))
-            + (self.nets_date_2 and '{self.nets_date_2:%d%m%y}' or ('0' * 6))
-            + (self.nets_date_3 and '{self.nets_date_3:%d%m%y}' or ('0' * 6))
+            f'{self.num_transactions:08d}'
+            f'{self.num_records:08d}'
+            + (self.total_amount and f'{self.total_amount:017d}' or ('0' * 17))
+            + (self.nets_date_1 and f'{self.nets_date_1:%d%m%y}' or ('0' * 6))
+            + (self.nets_date_2 and f'{self.nets_date_2:%d%m%y}' or ('0' * 6))
+            + (self.nets_date_3 and f'{self.nets_date_3:%d%m%y}' or ('0' * 6))
             + ('0' * 21)
-        ).format(self=self)
+        )
 
 
-@attr.s
-class TransactionRecord(Record):
-    transaction_type = attr.ib(converter=to_transaction_type)
-    transaction_number = attr.ib(converter=int)
+@define
+class TransactionRecord(Record, ABC):
+    """Transaction record base class."""
+
+    transaction_type: 'TransactionType' = field(converter=to_transaction_type)
+    transaction_number: int = field(converter=int)
 
 
-@attr.s
+@define
 class TransactionAmountItem1(TransactionRecord):
     """TransactionAmountItem1 is the first record of a transaction.
 
     The record is used both for AvtaleGiro and for OCR Giro.
     """
 
-    nets_date = attr.ib(converter=to_date)
-    amount = attr.ib(converter=int)
-    kid = attr.ib(
+    nets_date: 'datetime.date' = field(converter=to_date)
+    amount: int = field(converter=int)
+    kid: Optional[str] = field(
         converter=to_safe_str_or_none, validator=optional(str_of_max_length(25))
     )
 
     # Only OCR Giro
-    centre_id = attr.ib(default=None, validator=optional(str_of_length(2)))
-    day_code = attr.ib(default=None, converter=value_or_none(int))
-    partial_settlement_number = attr.ib(
-        default=None, converter=value_or_none(int)
-    )
-    partial_settlement_serial_number = attr.ib(
+    centre_id: Optional[str] = field(default=None, validator=optional(str_of_length(2)))
+    day_code: Optional[int] = field(default=None, converter=to_int_or_none)
+    partial_settlement_number: Optional[int] = field(default=None, converter=to_int_or_none)
+    partial_settlement_serial_number: Optional[str] = field(
         default=None, validator=optional(str_of_length(5))
     )
-    sign = attr.ib(default=None, validator=optional(str_of_length(1)))
+    sign: Optional[str] = field(default=None, validator=optional(str_of_length(1)))
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_1
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AMOUNT_ITEM_1
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -483,29 +447,31 @@ class TransactionAmountItem1(TransactionRecord):
 
     def to_ocr(self) -> str:
         """Get record as OCR string."""
-        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
+        if self.service_code == ServiceCode.OCR_GIRO:
             ocr_giro_fields = (
-                '{self.centre_id:2}'
-                '{self.day_code:02d}'
-                '{self.partial_settlement_number:01d}'
-                '{self.partial_settlement_serial_number:5}'
-                '{self.sign:1}'
-            ).format(self=self)
+                f'{self.centre_id:2}'
+                f'{self.day_code:02d}'
+                f'{self.partial_settlement_number:01d}'
+                f'{self.partial_settlement_serial_number:5}'
+                f'{self.sign:1}'
+            )
         else:
             ocr_giro_fields = ' ' * 11
 
         return (
             'NY'
-            '{self.service_code:02d}'
-            '{self.transaction_type:02d}'
+            f'{self.service_code:02d}'
+            f'{self.transaction_type:02d}'
             '30'
-            '{self.transaction_number:07d}'
-            '{self.nets_date:%d%m%y}' + ocr_giro_fields + '{self.amount:017d}'
-            '{self.kid:>25}' + ('0' * 6)
-        ).format(self=self)
+            f'{self.transaction_number:07d}'
+            f'{self.nets_date:%d%m%y}'
+            + ocr_giro_fields
+            + f'{self.amount:017d}{self.kid:>25}'
+            + ('0' * 6)
+        )
 
 
-@attr.s
+@define
 class TransactionAmountItem2(TransactionRecord):
     """TransactionAmountItem2 is the second record of a transaction.
 
@@ -513,21 +479,21 @@ class TransactionAmountItem2(TransactionRecord):
     """
 
     # TODO Validate `reference` length, which depends on service code
-    reference = attr.ib(converter=to_safe_str_or_none)
+    reference: Optional[str] = field(converter=to_safe_str_or_none)
 
     # Only OCR Giro
-    form_number = attr.ib(default=None, validator=optional(str_of_length(10)))
-    bank_date = attr.ib(default=None, converter=to_date)
-    debit_account = attr.ib(default=None, validator=optional(str_of_length(11)))
+    form_number: Optional[str] = field(default=None, validator=optional(str_of_length(10)))
+    bank_date: Optional['datetime.date'] = field(default=None, converter=to_date_or_none)
+    debit_account: Optional[str] = field(default=None, validator=optional(str_of_length(11)))
     # XXX In use in OCR Giro "from giro debited account" transactions in test
     # data, but documented as a filler field.
-    _filler = attr.ib(default=None)
+    _filler: Optional[str] = field(default=None)
 
     # Only AvtaleGiro
-    payer_name = attr.ib(default=None, converter=to_safe_str_or_none)
+    payer_name: Optional[str] = field(default=None, converter=to_safe_str_or_none)
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_2
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AMOUNT_ITEM_2
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -575,52 +541,43 @@ class TransactionAmountItem2(TransactionRecord):
     def to_ocr(self) -> str:
         """Get record as OCR string."""
         common_fields = (
-            'NY'
-            '{self.service_code:02d}'
-            '{self.transaction_type:02d}'
-            '31'
-            '{self.transaction_number:07d}'
-        ).format(self=self)
-
-        if self.service_code == netsgiro.ServiceCode.OCR_GIRO:
+            f'NY{self.service_code:02d}{self.transaction_type:02d}31{self.transaction_number:07d}'
+        )
+        if self.service_code == ServiceCode.OCR_GIRO:
             service_fields = (
-                '{self.form_number:10}'
-                + (self.reference and '{self.reference:9}' or (' ' * 9))
-                + (self._filler and '{self._filler:7}' or ('0' * 7))
-                + (self.bank_date and '{self.bank_date:%d%m%y}' or '0' * 6)
-                + '{self.debit_account:11}'
+                f'{self.form_number:10}'
+                + (self.reference and f'{self.reference:9}' or (' ' * 9))
+                + (self._filler and f'{self._filler:7}' or ('0' * 7))
+                + (self.bank_date and f'{self.bank_date:%d%m%y}' or '0' * 6)
+                + f'{self.debit_account:11}'
                 + ('0' * 22)
-            ).format(self=self)
-        elif self.service_code == netsgiro.ServiceCode.AVTALEGIRO:
+            )
+        elif self.service_code == ServiceCode.AVTALEGIRO:
             service_fields = (
-                (
-                    self.payer_name
-                    and '{:10}'.format(self.payer_name[:10])
-                    or (' ' * 10)
-                )
+                (self.payer_name and f'{self.payer_name[:10]:10}' or (' ' * 10))
                 + (' ' * 25)
-                + (self.reference and '{self.reference:25}' or (' ' * 25))
+                + (self.reference and f'{self.reference:25}' or (' ' * 25))
                 + ('0' * 5)
-            ).format(self=self)
-        else:
+            )
+        else:  # pragma: no cover
             service_fields = ' ' * 35
 
         return common_fields + service_fields
 
 
-@attr.s
+@define
 class TransactionAmountItem3(TransactionRecord):
     """TransactionAmountItem3 is the third record of a transaction.
 
     The record is only used for some OCR Giro transaction types.
     """
 
-    text = attr.ib(
+    text: Optional[str] = field(
         converter=to_safe_str_or_none, validator=optional(str_of_max_length(40))
     )
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AMOUNT_ITEM_3
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AMOUNT_ITEM_3
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -642,16 +599,13 @@ class TransactionAmountItem3(TransactionRecord):
     def to_ocr(self) -> str:
         """Get record as OCR string."""
         return (
-            'NY09'
-            '{self.transaction_type:02d}'
-            '32'
-            '{self.transaction_number:07d}'
-            + (self.text and '{:40}'.format(self.text) or (' ' * 40))
+            f'NY09{self.transaction_type:02d}32{self.transaction_number:07d}'
+            + (self.text and f'{self.text:40}' or (' ' * 40))
             + ('0' * 25)
-        ).format(self=self)
+        )
 
 
-@attr.s
+@define
 class TransactionSpecification(TransactionRecord):
     """TransactionSpecification is used for AvtaleGiro transactions.
 
@@ -662,15 +616,15 @@ class TransactionSpecification(TransactionRecord):
     specification text.
     """
 
-    line_number = attr.ib(converter=int)
-    column_number = attr.ib(converter=int)
-    text = attr.ib(
-        converter=stripped_newlines(fixed_len_str(40, str)),
+    line_number: int = field(converter=int)
+    column_number: int = field(converter=int)
+    text = field(
+        converter=stripped_newlines(fixed_len_str(40, str)),  # type: ignore[misc]
         validator=optional(str_of_max_length(40)),
     )
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_SPECIFICATION
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_SPECIFICATION
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -699,63 +653,58 @@ class TransactionSpecification(TransactionRecord):
 
     @classmethod
     def from_text(
-        cls, *, service_code, transaction_type, transaction_number, text
+        cls,
+        *,
+        service_code: ServiceCode,
+        transaction_type: 'TransactionType',
+        transaction_number: int,
+        text: Optional[str],
     ) -> Iterable['TransactionSpecification']:
         """Create a sequence of specification records from a text string."""
-
-        for line, column, text in cls._split_text_to_lines_and_columns(text):
+        for line, column, txt in cls._split_text_to_lines_and_columns(text or ''):
             yield cls(
                 service_code=service_code,
                 transaction_type=transaction_type,
                 transaction_number=transaction_number,
                 line_number=line,
                 column_number=column,
-                text=text,
+                text=txt,
             )
 
     @classmethod
-    def _split_text_to_lines_and_columns(
-        cls, text
-    ) -> Iterable[Tuple[int, int, str]]:
+    def _split_text_to_lines_and_columns(cls, text: str) -> Iterable[Tuple[int, int, str]]:
         lines = text.splitlines()
 
         if len(lines) > cls._MAX_LINES:
-            raise ValueError(
-                'Max {} specification lines allowed, got {}'.format(
-                    cls._MAX_LINES, len(lines)
-                )
-            )
+            raise ValueError(f'Max {cls._MAX_LINES} specification lines allowed, got {len(lines)}')
 
         for line_number, line_text in enumerate(lines, 1):
             if len(line_text) > cls._MAX_LINE_LENGTH:
                 raise ValueError(
-                    'Specification lines must be max {} chars long, '
-                    'got {}: {!r}'.format(
+                    'Specification lines must be max {} chars long, got {}: {!r}'.format(
                         cls._MAX_LINE_LENGTH, len(line_text), line_text
                     )
                 )
 
-            yield (line_number, 1, '{:40}'.format(line_text[0:40]))
-            yield (line_number, 2, '{:40}'.format(line_text[40:80]))
+            yield line_number, 1, f'{line_text[:40]:40}'
+            yield line_number, 2, f'{line_text[40:80]:40}'
 
     @classmethod
-    def to_text(cls, records: Sequence['TransactionSpecification']) -> str:
+    def to_text(cls, records: List['TransactionSpecification']) -> str:
         """Get a text string from a sequence of specification records."""
-
         if len(records) > cls._MAX_RECORDS:
             raise ValueError(
-                'Max {} specification records allowed, got {}'.format(
-                    cls._MAX_RECORDS, len(records)
-                )
+                f'Max {cls._MAX_RECORDS} specification records allowed, got {len(records)}'
             )
 
-        tuples = sorted([(r.line_number, r.column_number, r) for r in records])
+        tuples = sorted((r.line_number, r.column_number, r) for r in records)
 
         text = ''
         for _, column, specification in tuples:
-            text += specification.text
-            if column == cls._MAX_COLUMNS:
-                text += '\n'
+            if specification.text:
+                text += specification.text
+                if column == cls._MAX_COLUMNS:
+                    text += '\n'
 
         return text
 
@@ -763,15 +712,16 @@ class TransactionSpecification(TransactionRecord):
         """Get record as OCR string."""
         return (
             'NY212149'
-            '{self.transaction_number:07d}'
+            f'{self.transaction_number:07d}'
             '4'
-            '{self.line_number:03d}'
-            '{self.column_number:01d}'
-            '{self.text:40}' + ('0' * 20)
-        ).format(self=self)
+            f'{self.line_number:03d}'
+            f'{self.column_number:01d}'
+            f'{self.text:40}'
+            + ('0' * 20)
+        )
 
 
-@attr.s
+@define
 class AvtaleGiroAgreement(TransactionRecord):
     """AvtaleGiroAgreement is used by Nets to notify about agreement changes.
 
@@ -779,14 +729,16 @@ class AvtaleGiroAgreement(TransactionRecord):
     notification preferences.
     """
 
-    registration_type = attr.ib(converter=to_avtalegiro_registration_type)
-    kid = attr.ib(
+    registration_type: 'AvtaleGiroRegistrationType' = field(
+        converter=to_avtalegiro_registration_type
+    )
+    kid: Optional[str] = field(
         converter=to_safe_str_or_none, validator=optional(str_of_max_length(25))
     )
-    notify = attr.ib(converter=to_bool)
+    notify: bool = field(converter=to_bool)
 
-    RECORD_TYPE = netsgiro.RecordType.TRANSACTION_AGREEMENTS
-    _PATTERNS = [
+    RECORD_TYPE: ClassVar[RecordType] = RecordType.TRANSACTION_AGREEMENTS
+    _PATTERNS: ClassVar[List[Pattern]] = [
         re.compile(
             r'''
             ^
@@ -810,30 +762,27 @@ class AvtaleGiroAgreement(TransactionRecord):
     def to_ocr(self) -> str:
         """Get record as OCR string."""
         return (
-            'NY219470'
-            '{self.transaction_number:07d}'
-            '{self.registration_type:01d}'
-            '{self.kid:>25}' + (self.notify and 'J' or 'N') + ('0' * 38)
+            f'NY219470{self.transaction_number:07d}{self.registration_type:01d}{self.kid:>25}'
+            + (self.notify and 'J' or 'N')
+            + ('0' * 38)
         ).format(self=self)
 
 
-def parse(data: str) -> List[Record]:
+def parse(data: str) -> List[R]:
     """Parse an OCR file into a list of record objects."""
 
-    def all_subclasses(cls):
-        return cls.__subclasses__() + [
-            subsubcls
-            for subcls in cls.__subclasses__()
-            for subsubcls in all_subclasses(subcls)
+    def all_subclasses(cls: Union[Type[R], Type[Record]]) -> List[Type[R]]:
+        """Return a list of subclasses for a given class."""
+        classes = cls.__subclasses__() + [
+            subsubcls for subcls in cls.__subclasses__() for subsubcls in all_subclasses(subcls)
         ]
+        return cast(List[Type[R]], classes)
 
     record_classes = {
-        cls.RECORD_TYPE: cls
-        for cls in all_subclasses(Record)
-        if hasattr(cls, 'RECORD_TYPE')
+        cls.RECORD_TYPE: cls for cls in all_subclasses(Record) if hasattr(cls, 'RECORD_TYPE')
     }
 
-    results = []
+    results: List[R] = []
 
     for line in data.strip().splitlines():
         if len(line) != 80:
@@ -841,11 +790,9 @@ def parse(data: str) -> List[Record]:
 
         record_type_str = line[6:8]
         if not record_type_str.isnumeric():
-            raise ValueError(
-                'Record type must be numeric, got {!r}'.format(record_type_str)
-            )
+            raise ValueError(f'Record type must be numeric, got {record_type_str!r}')
 
-        record_type = netsgiro.RecordType(int(record_type_str))
+        record_type = to_record_type(record_type_str)
         record_cls = record_classes[record_type]
 
         results.append(record_cls.from_string(line))
